/// @addtogroup covrel
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file HSet.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_COVREL_HSET_H_ 
#define _CAPD_COVREL_HSET_H_ 

#include <vector>

namespace capd{
namespace covrel{

/**
   This class is used to store a grid of a h-set in the form
   center[i] + M*r
*/
template<typename MatrixT>
class GridSet
{
public:
  typedef MatrixT MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef typename std::vector<VectorType>::iterator iterator;
  typedef typename std::vector<VectorType>::const_iterator const_iterator;
   
  iterator begin(){
    return center.begin();
  }
   
  iterator end(){
    return center.end();
  }

  const_iterator begin() const {
    return center.begin();
  }
   
  const_iterator end() const {
    return center.end();
  }

  std::vector<VectorType> center;
  MatrixType C;
  VectorType r;
};

/**
   this is an abstract class for h-sets - from paper by Gidea-Zgliczynski
   'Covering relations ...'  http://www.im.uj.edu.pl/~zgliczyn
*/

template <typename VectorType, typename IVectorType>
class HSet{
public:

  //  methods for nonrigorous computations - used usually in tests and setting up parameters
  virtual bool outside(const VectorType &) const = 0; // verifies if a vector does not intersect support of h-set
  virtual bool inside (const VectorType &) const = 0; // verifies if a vector is inside the support of h-set
  virtual bool across (const VectorType &) const = 0; // verifies if a vector does not intersect N^+
  virtual bool mapaway(const VectorType &) const = 0; // verifies if the projection onto stable directions
                                                         // of a vector is outside the unit ball in a proper
                                                         // coordinate system

  // interval (rigorous) versions:
  virtual bool outside(const IVectorType &) const = 0;
  virtual bool inside (const IVectorType &) const = 0;
  virtual bool across (const IVectorType &) const = 0;
  virtual bool mapaway(const IVectorType &) const = 0;

   virtual ~HSet(){}
}; //the end of abstract class HSet

}} // namespace capd::covrel

#endif // _CAPD_COVREL_HSET_H_ 

/// @}
