/// @addtogroup diffIncl
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file DiffInclusionCW.hpp
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Tomasz Kapela, 2007 */

#ifndef _CAPD_DIFFINCL_DIFFINCLUSIONCW_HPP_ 
#define _CAPD_DIFFINCL_DIFFINCLUSIONCW_HPP_ 

#include <sstream>
#include <string>
#include <stdexcept>

#include "diffIncl/DiffInclusionCW.h"
#include "diffIncl/DiffInclusion.hpp"

namespace capd{
namespace diffIncl{


template<typename MapT>
DiffInclusionCW<MapT>::DiffInclusionCW(
           MultiMapType& A_diffIncl, 
           int A_order, 
           ScalarType const & A_step, 
           NormType const & A_norm
) 
  : capd::diffIncl::DiffInclusion<MapT>(A_diffIncl, A_order, A_step, A_norm) {
}

/// Component wise algorithm
template<typename MapType>
typename DiffInclusionCW<MapType>::VectorType DiffInclusionCW<MapType>::perturbations(const VectorType& x){
  
  VectorType W_1 = dynamicalSystemEnclosure(x);
  VectorType W_2 = diffInclusionEnclosure(x);
  
  MatrixType J = m_diffIncl[W_2]; 
  VectorType deltha = m_diffIncl.perturbations(W_1);
  VectorType C = rightVector(deltha);
  
  int i, j;
  for( i=0; i < J.numberOfRows(); ++i)
    for( j = 0; j < J.numberOfColumns(); ++j)
      if(i != j)
        J[i][j] = (capd::abs(J[i][j])).right();
      else
        J[i][j] = (J[i][j]).right();
  MatrixType At = getStep() * J;
  MatrixType A = MatrixType::Identity(J.numberOfRows());
  MatrixType Sum = A;
  ScalarType AtNorm = right((*m_norm)(At));
  
  // an expected error bound for remainder
  ScalarType epsilon = 1.0 / 1024.0;
  // we assume |A_N| = 1 then from expresion for remainder (see below)
  int N = max(static_cast<int>( (AtNorm *(1/epsilon + 1)).rightBound()) -2, 1);
  
  for(i= 0; i < N; ++i){
    A = A * At / ScalarType(i + 2.0);
    Sum += A;  
  }
  // remainder = |A| * |At/(N+2)| / (1 - |At/(N+2)|)   (the sum of geometric series)
  ScalarType remainder = ((*m_norm)(A) * AtNorm / (N + 2 - AtNorm )).right();
  
  for(i=0; i < J.numberOfRows(); ++i)
    for(int j = 0; j < J.numberOfColumns(); ++j)
      Sum[i][j] += remainder * ScalarType(-1.0, 1.0);
  VectorType D = getStep() * (Sum * C);
  VectorType result(D.dimension());
    
  for(int i=0; i< D.dimension(); ++i)
    result[i] = ScalarType(-D[i].rightBound(), D[i].rightBound());
  
  return result;
}
  
}} //namespace capd::diffIncl

#endif // _CAPD_DIFFINCL_DIFFINCLUSIONCW_HPP_ 

/// @}
