/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file C1Rect.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_DYNSET_C1RECT_HPP_ 
#define _CAPD_DYNSET_C1RECT_HPP_ 

#include <sstream>
#include "dynset/C1Rect.h"
#include "capd/minmax.h"
#include "matrixAlgorithms/floatMatrixAlgorithms.hpp"

namespace capd{
namespace dynset{

template<typename MatrixType>
C1Rect<MatrixType>::C1Rect(const NormType& aNorm, int dim)
  : m_x(dim),
    m_r(dim),
    m_B(MatrixType::Identity(dim)),
    m_D(MatrixType::Identity(dim)),
    m_Bjac(MatrixType::Identity(dim)),
    m_R(dim,dim),
    m_Norm(aNorm.clone())
{
  m_x.clear();
  m_r.clear();
  m_R.clear();
}

template<typename MatrixType>
C1Rect<MatrixType>::C1Rect(const C1Rect<MatrixType>& s)
  : m_x(s.m_x),
    m_r(s.m_r),
    m_B(s.m_B),
    m_D(s.m_D),
    m_Bjac(s.m_Bjac),
    m_R(s.m_R),
    m_Norm(s.m_Norm->clone())
{}

template<typename MatrixType>
C1Rect<MatrixType>::C1Rect(const VectorType& x, const NormType& aNorm)
  : m_x(x),
    m_r(x.dimension()),
    m_B(MatrixType::Identity(x.dimension())),
    m_D(MatrixType::Identity(x.dimension())),
    m_Bjac(MatrixType::Identity(x.dimension())),
    m_R(x.dimension(),x.dimension()),
    m_Norm(aNorm.clone())
{
  split(m_x,m_r);
  m_R.clear();
}

template<typename MatrixType>
C1Rect<MatrixType>::C1Rect(const VectorType& x,const VectorType& r,const NormType& theNorm)
  : m_x(x),
    m_r(r),
    m_B(MatrixType::Identity(x.dimension())),
    m_D(MatrixType::Identity(x.dimension())),
    m_Bjac(MatrixType::Identity(x.dimension())),
    m_R(x.dimension(),x.dimension()),
    m_Norm(theNorm.clone())
{
  m_R.clear();
// the following assures that m_r contains zero
  if( !subset(VectorType(x.dimension()),m_r) )
  {
    m_x+=m_r;
    split(m_x,m_r);
  }
}


template<typename MatrixType>
C1Rect<MatrixType>::C1Rect(
    const VectorType& x,
    const MatrixType& B,
    const VectorType& r,
    const NormType& theNorm
  ):m_x(x),
    m_r(r),
    m_B(B),
    m_D(MatrixType::Identity(x.dimension())),
    m_Bjac(MatrixType::Identity(x.dimension())),
    m_R(x.dimension(),x.dimension()),
    m_Norm(theNorm.clone())
{
  m_R.clear();
// the following assures that m_r contains zero and the representation is valid
  if( !subset(VectorType(x.dimension()),m_r) )
  {
    VectorType centerR = m_r;
    split(centerR,m_r);
    // m_x will be not a single point, but we must assure m_r contains zero
    m_x += m_B*centerR; 
  }
}

template<typename MatrixType>
void C1Rect<MatrixType>::move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys, C1Rect& result) const
{
  int dim = m_x.dimension();

  VectorType Rem(dim);
  MatrixType jacRem(dim,dim);

  VectorType xx = m_x + m_B*m_r;

  int found;
  VectorType z = c1dynsys.enclosure(xx,&found);
  MatrixType Z = c1dynsys.jacEnclosure(z,*m_Norm);
  c1dynsys.JacRemainder(z,Z,Rem,jacRem);
  
  MatrixType Jac = c1dynsys.JacPhi(xx);
  result.m_x = c1dynsys.Phi(m_x) + Rem;
// we use again z, the current value is unnecessary now
  split(result.m_x,z);

  Z = Jac*m_B;
  mid(Z,result.m_B);
  capd::matrixAlgorithms::orthonormalize(result.m_B);
  MatrixType Qtr = Transpose(result.m_B);
  result.m_r = (Qtr*Z)*m_r+Qtr*z;

// evaluation formula for C^1 part
// we evaluate the multiplication (Jac+jacRem)*(D+B*R)
  Jac += jacRem;

// we use jacRem again
  split(Jac,jacRem);

  result.m_D = Jac*m_D + jacRem*(m_D + m_Bjac * m_R);
  split(result.m_D,Z);
// we use jacRem again
  jacRem = Jac*m_Bjac;
  mid(jacRem,result.m_Bjac);
  capd::matrixAlgorithms::orthonormalize(result.m_Bjac);
  Qtr = Transpose(result.m_Bjac);
  result.m_R = (Qtr*jacRem)*m_R + Qtr*Z;
}

// ----------------------------------------------

template<typename MatrixType>
typename C1Rect<MatrixType>::ScalarType C1Rect<MatrixType>::size(void) const
{
  return capd::vectalg::size(m_x + m_B*m_r);
}

template<typename MatrixType>
std::string C1Rect<MatrixType>::show(void) const
{
  std::ostringstream descriptor;
  descriptor << "C1Rect:\n x=";
  descriptor << m_x << "\n B=";
  descriptor << m_B << "\n r=";
  descriptor << m_r << "\n D= ";
  descriptor << m_D << "\n Bjac= ";
  descriptor << m_Bjac << "\n R= ";
  descriptor << m_R << '\0';

  return descriptor.str();
}

template<typename MatrixT>
C1Rect<MatrixT>::operator typename C1Rect<MatrixT>::MatrixType() const
{
  return m_D + m_Bjac*m_R;
}

template<typename MatrixType>
C1Rect<MatrixType> &C1Rect<MatrixType>::operator=(const VectorType& x)
{
  int dim = x.dimension();
  m_x = x;
  m_r = VectorType(dim);
  split(m_x,m_r);

  m_Bjac = m_D = m_B = MatrixType::Identity(dim);
  m_R = MatrixType(dim,dim);
  m_R.clear();

  return *this;
}

template<typename MatrixType>
C1Rect<MatrixType>& C1Rect<MatrixType>::operator=(const C1Rect& S)
{
  m_x = S.m_x;
  m_B = S.m_B;
  m_r = S.m_r;

  m_D = S.m_D;
  m_Bjac = S.m_Bjac;
  m_R = S.m_R;
  delete m_Norm;
  m_Norm = S.m_Norm->clone();
  return *this;
}

}} // namespace capd::dynset

#endif // _CAPD_DYNSET_C1RECT_HPP_ 

/// @}
