/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file BasicTaylor.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Daniel Wilczak 2001-2004 */

#ifndef _CAPD_DYNSYS_BASICTAYLOR_HPP_ 
#define _CAPD_DYNSYS_BASICTAYLOR_HPP_ 

#include <sstream>
#include <string>
#include <stdexcept>

#include "capd/factrial.h"
#include "dynsys/BasicTaylor.h"

namespace capd{
namespace dynsys{

//###########################################################//

template<typename MapType>
BasicTaylor<MapType>::BasicTaylor(
      MapType& vectorField,
      int a_order,
      const ScalarType& Step
  ) : m_vField(&vectorField), m_order(a_order), m_step(Step)
{
  m_dimension = m_vField->getDimension();
  m_vField->setOrder(m_order+1);
  m_coeff = new (m_dimension) VectorType[m_order+2];
  m_matrixCoeff = new (m_dimension,m_dimension) MatrixType[m_order+2];
  m_F = new (m_dimension,m_dimension) MatrixType[m_order+2];
}


//###########################################################//


template<typename MapType>
BasicTaylor<MapType>::~BasicTaylor()
{
  delete [] m_coeff;
  delete [] m_matrixCoeff;
  delete [] m_F;
}

//###########################################################//


template<typename MapType>
void BasicTaylor<MapType>::setOrder(int newOrder)
{
  if(newOrder!=m_order)
  {
    m_order = newOrder;
    m_vField->setOrder(m_order+1);

    delete [] m_coeff;
    delete [] m_matrixCoeff;
    delete [] m_F;
    m_coeff = new (m_dimension) VectorType[m_order+2];
    m_matrixCoeff = new (m_dimension,m_dimension) MatrixType[m_order+2];
    m_F = new (m_dimension,m_dimension) MatrixType[m_order+2];
  }
}


//###########################################################//


template<typename MapType>
void BasicTaylor<MapType>::computeMatrixCoeff(int order, const MatrixType &V) const
{
  if(order<0)
  {
    std::ostringstream message;
    message << "Taylor error: negative index " << order << " of Taylor coefficient in function computeMatrixCoeff";
    throw std::out_of_range(message.str());
  }

  // now we compute the coefficients for the Jacobian matrix
  // the coefficients will be stored in F
  m_vField->variational(m_coeff,m_F,order+1);

  m_matrixCoeff[0] = V;

  int p,i,j,k,r;
  for(p=0;p<=order;++p)
  {
    typename MatrixType::iterator b = m_matrixCoeff[p+1].begin();
    for(i=1;i<=m_dimension;++i)
      for(j=1;j<=m_dimension;++j)
      {
        ScalarType result(0);
        for(k=1;k<=m_dimension;++k)
          for(r=0;r<=p;++r)
            result += m_F[r](i,k) * m_matrixCoeff[p-r](k,j);
          *b = result/(p+1);
          ++b;
      }
  }
}


//###########################################################//


template<typename MapType>
typename BasicTaylor<MapType>::VectorType BasicTaylor<MapType>::operator()(const VectorType &iv) const
{
  m_coeff[0] = iv;   //  coeff will contain  Taylor coefficients in <iv>
                      // now we initiate 0-th order term

  int r;
  for(r=0;r<m_order;++r)
    m_coeff[r+1] = (*m_vField)(m_coeff[r],r)/(r+1);  // recurrent computation of consequtive
                                             // Taylor coefficients
  VectorType result = m_coeff[m_order];
  for(r=m_order-1;r>=0;--r)
    result = result*m_step + m_coeff[r];
  return result;
}


//###########################################################//


template<typename MapType>
typename BasicTaylor<MapType>::VectorType BasicTaylor<MapType>::operator()(
      const VectorType &iv, MatrixType& der
  ) const
{
  // first we compute ceofficients for x
  int r;
  m_coeff[0] = iv;
  for(r=0;r<m_order;++r)
    m_coeff[r+1] = (*m_vField)(m_coeff[r],r)/(r+1);

  VectorType result = m_coeff[m_order];
  for(r=m_order-1;r>=0;--r)
    result = result*m_step + m_coeff[r];

  // this function compute coefficients for variational equation up to index 'order'
  computeMatrixCoeff(m_order-1,MatrixType::Identity(m_dimension));

  // the sumation of the Taylor series
  der = m_matrixCoeff[m_order];
  for(r=m_order-1;r>=0;--r)
    der = der*m_step + m_matrixCoeff[r];

  return result;
}


}} //namespace capd::dynsys

#endif // _CAPD_DYNSYS_BASICTAYLOR_HPP_ 

/// @}
