/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file CnTaylor.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Daniel Wilczak 2006 */

#ifndef _CAPD_DYNSYS_CNTAYLOR_H_ 
#define _CAPD_DYNSYS_CNTAYLOR_H_ 

#include <string>
#include <vector>
#include "vectalg/Norm.h"
#include "dynsys/C2Taylor.h"
#include "dynsys/CnDynSys.h"
#include "dynsys/BasicCnTaylor.h"

namespace capd{
namespace dynsys{

template<typename MapT>
class CnTaylor : public BasicCnTaylor<MapT>, public CnDynSys<typename MapT::MatrixType>{
public:
  typedef MapT MapType;
  typedef MapT CnMapType;
  typedef typename MapType::FunctionType FunctionType;
  typedef typename CnMapType::MatrixType MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef typename CnMapType::C2CoeffType C2CoeffType;
  typedef typename CnMapType::CnCoeffType CnCoeffType;
  typedef typename CnMapType::NodeType NodeType;
  typedef typename CnMapType::ConsNodeType ConsNodeType;
  typedef typename CnMapType::SeriesContainerType SeriesContainerType;
  typedef typename SeriesContainerType::Multipointer Multipointer;
  typedef typename SeriesContainerType::Multiindex Multiindex;
  typedef capd::vectalg::Norm<VectorType,MatrixType> NormType;

  CnTaylor(MapType& vectorField, int order, const ScalarType& step);

// enclosure
  VectorType enclosure(const VectorType& x, int* found) const;
  VectorType enclosure(const VectorType& x) const;
  MatrixType jacEnclosure(const VectorType& enc, const NormType& norm) const;
  void c2Enclosure(
         const VectorType& W1,
         const MatrixType& W3,
         const NormType& the_norm,
         C2CoeffType& result
  ) const;
  void cnEnclosure(const VectorType& x, CnCoeffType& result, const NormType& the_norm) const;
  VectorType cnEnclosure(const VectorType& x, const NormType& the_norm, int rank) const;

// Phi
  VectorType Phi(const VectorType &iv) const;
  MatrixType JacPhi(const VectorType &iv) const;
  void c2Phi(const VectorType& x, MatrixType& jac, C2CoeffType& hessian) const;
  void cnPhi(const VectorType&x, CnCoeffType& result) const;

// remainder
  VectorType Remainder(const VectorType &iv) const;
  void JacRemainder(
            const VectorType &vecEnclosure,
            const MatrixType &jacEnclosure,
            VectorType &Remainder,
            MatrixType &jRemainder
    ) const;
  void c2Remainder(
         const VectorType& Enclosure,
         const MatrixType& jacEnclosure,
         const C2CoeffType& hessianEnclosure,
         VectorType& Remainder,
         MatrixType& jacRemainder,
         C2CoeffType& hessianRemainder // here a result is stored
    ) const;
  void cnRemainder(const VectorType& x, const NormType&, CnCoeffType& result) const;

  using DynSys<MatrixType>::Lipschitz;
  using DynSys<MatrixType>::type;
  using BasicCnTaylor<MapType>::getDimension;
  using BasicCnTaylor<MapType>::getRank;

protected:
  void operator=(const CnTaylor&){}
  CnTaylor(const CnTaylor& t) : BasicCnTaylor<MapT>(t){}

  using BasicCnTaylor<MapType>::computeNonlinearPart;
  using BasicCnTaylor<MapType>::m_order;
  using BasicCnTaylor<MapType>::m_step;
  using BasicCnTaylor<MapType>::m_vectorField;
  using BasicCnTaylor<MapType>::m_vectorFieldSeries;
  using BasicCnTaylor<MapType>::m_resultSeries;
  using BasicCnTaylor<MapType>::m_compositionFormula;
  using BasicCnTaylor<MapType>::m_nonlinearPart;
  using BasicCnTaylor<MapType>::computeVectorFieldSeries;
  using BasicCnTaylor<MapType>::computeResultSeries;
   
  ScalarType computeDelta(const Multipointer&) const;
}; // the end of class CnTaylor

// ---------------------------- CONSTRUCTORS ---------------------------------------------

template<typename CnMapT>
inline CnTaylor<CnMapT>::CnTaylor(CnMapType& a_vectorField, int a_order, const ScalarType& a_step)
  : BasicCnTaylor<CnMapT>(a_vectorField,a_order,a_step)
{}


template<typename MapType>
inline typename CnTaylor<MapType>::VectorType CnTaylor<MapType>::enclosure(const VectorType &x, int *found) const
// the function finds an enclosure for \varphi([0,step],x)
{
  return enclosure(x);
}

}} // the end of the namespace capd::dynsys

#endif // _CAPD_DYNSYS_CNTAYLOR_H_ 

/// @}
