/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file FadTaylor.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by the CAPD Group.
//
// Distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#ifndef _CAPD_DYNSYS_FADTAYLOR_H_ 
#define _CAPD_DYNSYS_FADTAYLOR_H_ 

#include "vectalg/Norm.hpp"
#include "dynsys/BasicFadTaylor.h"
#include "dynsys/C1DynSys.h"
#include "dynsys/enclosure.hpp"
#include "dynsys/DynSys.hpp"
#include "capd/power.h"

namespace capd{
namespace dynsys{

template<class FadMapT>
class FadTaylor : public capd::dynsys::C1DynSys<typename FadMapT::MatrixType>, public BasicFadTaylor<FadMapT>{
public:
  typedef FadMapT MapType;
  typedef typename FadMapT::ScalarType ScalarType;
  typedef typename FadMapT::FScalar FScalar;
  typedef typename FadMapT::TFScalar TFScalar;
  typedef typename FadMapT::MatrixType MatrixType;
  typedef typename FadMapT::VectorType VectorType;
  typedef typename FadMapT::FVector FVector;
  typedef typename FadMapT::TFVector TFVector;
  typedef typename MapType::FunctionType FunctionType;

  typedef Norm<VectorType,MatrixType> NormType;
  typedef BasicFadTaylor<FadMapT> BasicFadTaylorType;

  FadTaylor(MapType& f, int _order, ScalarType _step)
  : BasicFadTaylorType(f,_order,_step)
  {}

// DynSys and C1DynSys interface implementation
      
  VectorType Phi(const VectorType& v) const{       ///< Computes image of vector v after one time step
    return BasicFadTaylorType::operator()(v);
  }

  MatrixType JacPhi(const VectorType& v) const{    ///< Computes derivatives with respect to initial conditions
    MatrixType der;
    BasicFadTaylorType::operator()(v,der);
    return der;
  }

  VectorType Remainder(const VectorType& v) const{  ///< Computes bound for remainder term in Taylor series
    this->operator()(enclosure(v),this->m_order+1);
    VectorType result(v.dimension());
    ScalarType fac = power(this->m_step,this->m_order+1);
    for(int i=0;i<v.dimension();++i)
      result[i] = fac*this->m_in[i][this->m_order+1].x();
    return result;
  }

  void JacRemainder(                               ///< Computes bounds for remainder term for Taylor series of derivative
          const VectorType &vecEnclosure,
          const MatrixType &jacEnclosure,
          VectorType &Remainder,
          MatrixType &jRemainder
  ) const{
    MatrixType der;
    BasicFadTaylorType::operator()(vecEnclosure,jRemainder,this->m_order+1,jacEnclosure);
    ScalarType fac = power(this->m_step,this->m_order+1);
    for(int i=0;i<vecEnclosure.dimension();++i)
    {
      Remainder[i] = fac*this->m_in[i][this->m_order+1].x();
      for(int j=0;j<vecEnclosure.dimension();++j)
        jRemainder(i+1,j+1) = fac*this->m_in[i][this->m_order+1].d(j);
    }
  }

  VectorType enclosure(const VectorType& x, int* found) const{ ///< the function finds an enclosure for \varphi([0,step],x)
    return capd::dynsys::enclosure(this->m_vectorField, x, this->m_step);
  }

  VectorType enclosure(const VectorType &x) const{             ///< the function finds an enclosure for \varphi([0,step],x)
    return capd::dynsys::enclosure(this->m_vectorField, x, this->m_step);
  }

  MatrixType jacEnclosure(const VectorType& enc, const NormType& norm) const{ ///< the function finds enclosure for Jacobian matrix (variational part)
    return capd::dynsys::jacEnclosure(this->m_vectorField,this->m_step,enc,norm);
  }
// end of Dynsys and C1DynSys members
      
  using BasicFadTaylorType::getOrder;
  using BasicFadTaylorType::setOrder;
  
  using BasicFadTaylorType::getStep;
  using BasicFadTaylorType::setStep;
  
}; // the end of class FadTaylor


}} // the end of the namespace capd::dynsys

#endif // _CAPD_DYNSYS_FADTAYLOR_H_ 

/// @}
