/// @addtogroup dynsys2
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Taylor.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by the CAPD Group.
//
// Distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

/* Authors: Daniel Wilczak, 2001-2004, Tomasz Kapela, 2008 */

#ifndef _CAPD_DYNSYS2_TAYLOR_H_ 
#define _CAPD_DYNSYS2_TAYLOR_H_ 

#include <string>
#include "vectalg/Norm.h"
#include "dynsys/C1DynSys.h"
#include "dynsys2/BasicTaylor.h"
#include "dynsys/enclosure.h"


namespace capd{
  namespace dynsys2{
    
    /**
     *  Taylor class represents Dynamical System \f$ \varphi(t, x) \f$ for given ODE.
     *
     *  We fix time step t=h so in fact we have discrete dynamical system.
     *  We integrate ODE using Taylor method. We can compute rigorous estimates for:
     *    -  image of vector x after one time step : \f$ \varphi(h, x)  = Phi(x) + Remainder(x)\f$
     *    -  Jacobian matrix : \f$ \frac{\partial \varphi(h,x)}{\partial x}  = JacPhi(x) + JacRemainder(...)\f$
     *
     */
    
    template<typename MapT>
    class Taylor:   public capd::dynsys::C1DynSys<typename MapT::MatrixType>, public BasicTaylor<MapT> {
    public:
      typedef MapT MapType;
      typedef typename MapT::FunctionType FunctionType;
      typedef typename MapType::MatrixType MatrixType;
      typedef typename MatrixType::RowVectorType VectorType;
      typedef typename MatrixType::ScalarType ScalarType;
      typedef capd::vectalg::Norm<VectorType, MatrixType> NormType;
      
      /// Constructor
      Taylor(MapType& Field, int Order, const ScalarType& Step);
      
// DynSys interface implementation
      
      VectorType Phi(const VectorType& v) const;       ///< Computes image of vector v after one time step
      MatrixType JacPhi(const VectorType& v) const;    ///< Computes derivatives with respect to initial conditions
      VectorType Remainder(const VectorType& v) const; ///< Computes error bounds for Phi
      void JacRemainder(                               ///< Computes error bounds for JacPhi
              const VectorType &vecEnclosure,
              const MatrixType &jacEnclosure,
              VectorType &Remainder,
              MatrixType &jRemainder
      ) const;
// end of Dynsys members
      
      
      VectorType enclosure(const VectorType& x, int* found) const; ///< the function finds an enclosure for \varphi([0,step],x)
      VectorType enclosure(const VectorType &x) const;             ///< the function finds an enclosure for \varphi([0,step],x)
      MatrixType jacEnclosure(const VectorType& enc, NormType& norm) const; ///< the function finds enclosure for Jacobian matrix (variational part)
      
      using BasicTaylor<MapType>::getField;
      using BasicTaylor<MapType>::setField;
      
      using BasicTaylor<MapType>::getOrder;
      using BasicTaylor<MapType>::setOrder;
      
      using BasicTaylor<MapType>::getStep;
      using BasicTaylor<MapType>::setStep;
      
      using BasicTaylor<MapType>::getCoeff;
      using BasicTaylor<MapType>::getMatrixCoeff;
      
    protected:
      using BasicTaylor<MapType>::computeMatrixCoeff;
      using BasicTaylor<MapType>::m_vField;
      using BasicTaylor<MapType>::m_order;
      using BasicTaylor<MapType>::m_step;
      using BasicTaylor<MapType>::m_coeff;
      using BasicTaylor<MapType>::m_matrixCoeff;
      using BasicTaylor<MapType>::m_dimension;
      using BasicTaylor<MapType>::m_F;
    };
    
    
// --------------- inline definitions -----------------
    
//-----------------------------------------------------------//
    
    template<typename MapType>
    inline typename Taylor<MapType>::VectorType Taylor<MapType>::Phi(const VectorType& iv) const {
      return BasicTaylor<MapType>::operator()(iv);
    }
    
//-----------------------------------------------------------//
    template<typename MapType>
    typename Taylor<MapType>::VectorType Taylor<MapType>::enclosure(const VectorType & x) const
    /// the function finds an enclosure for \varphi([0,step],x)
    {
      return capd::dynsys::enclosure(m_vField, x, m_step);
    }
    
//-----------------------------------------------------------//
    /** 
     * the function finds an enclosure for \varphi([0,step],x) 
     *
     * @param[in]  x      set of initial conditions
     * @param[out] found  not used any more (in the past it returned information if eclosure was found
     */
    template<typename MapType>
    typename Taylor<MapType>::VectorType Taylor<MapType>::enclosure(const VectorType & x, int *found) const {
      return enclosure(x);
    }

    template<typename MapType>
    typename Taylor<MapType>::MatrixType Taylor<MapType>::jacEnclosure(const VectorType& enc, NormType& norm) const{
      return capd::dynsys::jacEnclosure(m_vField,m_step,enc, norm);
    }
    
    
  }} // namespace capd::dynsys

#endif // _CAPD_DYNSYS2_TAYLOR_H_ 

/// @}
