/// @addtogroup dynsys2
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Taylor.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by the CAPD Group.
//
// Distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

/* Authors: Daniel Wilczak, 2001-2004, Tomasz Kapela, 2008 */

#ifndef _CAPD_DYNSYS2_TAYLOR_HPP_ 
#define _CAPD_DYNSYS2_TAYLOR_HPP_ 

#include <sstream>
#include <string>
#include <stdexcept>

#include "capd/factrial.h"
#include "dynsys/TaylorException.h"
#include "dynsys2/Taylor.h"
#include "dynsys/DynSys.hpp"

namespace capd{
  namespace dynsys2{
    
//-----------------------------------------------------------//
    /**
     *  Constructor
     *
     *  @param[in] vectorField   vector field (right hand side of ODE)
     *  @param[in] order         order of Taylor expansion
     *  @param[in] step          time step for ODE integration
     *
     */
    template<typename MapType>
    Taylor<MapType>::Taylor(MapType& vectorField, int order, const ScalarType& step)
    : BasicTaylor<MapType>(vectorField, order, step)
    {}
    
    
//-----------------------------------------------------------//
    
    template<typename MapType>
    typename Taylor<MapType>::MatrixType Taylor<MapType>::JacPhi(const VectorType &iv) const {
      MatrixType result(m_dimension, m_dimension);
      (*this)(iv, result);
      return result;
    }
    
    
//-----------------------------------------------------------//
    
    
    template<typename MapType>
    typename Taylor<MapType>::VectorType Taylor<MapType>::Remainder(const VectorType &iv) const {
      // we compute an enclosure for \varphi([0,timestep],iv)
      m_coeff[0] = enclosure(iv);
      
// the computation of Taylor coefficients up to an order <order+1> on <enc>
      for(int i=0;i<=m_order;++i)
        m_coeff[i+1] = m_vField(m_coeff[i], i)/(i+1);
      
      return m_coeff[m_order+1]*power(m_step, m_order+1);
    }
    
    
//-----------------------------------------------------------//
    /**
     * Computes the remainder for varaiational part 
     * 
     * @param vecEnc  enclosure for \varphi([0,step],x)
     * @param jacEnc   enclosure for V([0,step],x)
     * @param rem 
     */
    template<typename MapType>
    void Taylor<MapType>::JacRemainder(
            const VectorType &vecEnc,
            const MatrixType &jacEnc,
            VectorType &rem,
            MatrixType &jacRem
    ) const
    {
      m_coeff[0] = vecEnc;
      int r;
      for(r=0;r<=m_order;++r)
        m_coeff[r+1] = m_vField(m_coeff[r], r)/(r+1);
      
     // m_vField.variational(m_coeff, m_F, m_order+1);
      computeMatrixCoeff(m_order, jacEnc);
      
      ScalarType fac = power(m_step, m_order+1);
      rem = m_coeff[m_order+1] * fac;
      jacRem = m_matrixCoeff[m_order+1] * fac;
    }
    
    
//-----------------------------------------------------------//
    
  }} //namespace capd::dynsys

#endif // _CAPD_DYNSYS2_TAYLOR_HPP_ 

/// @}
