/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file  PoincareMap.h
///
/// @author Daniel Wilczak
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by the CAPD Group.
//
// Distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

#ifndef _CAPD_POINCARE_POINCARE_MAP_HPP_
#define _CAPD_POINCARE_POINCARE_MAP_HPP_

#include "poincare/PoincareMap.h"
#include "poincare/BasicPoincareMap.hpp"

namespace capd{
namespace poincare{

template<typename DS, typename FunT>
double PoincareMap<DS, FunT>::timeStepCorrectionFactor = 0.9;
  
template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet)
{
// The originalSet after this function contains a set just ater section.
   derivativeOfFlow = NULL;
   hessianOfFlow = NULL;
   partialDerivatives = NULL;
// We move the set to be very close to the section   
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet);
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet, MatrixType& der)
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = &der;
   hessianOfFlow = NULL;
   partialDerivatives = NULL;
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet);
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet, MatrixType& der, MatrixType hessian[])
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = &der;
   hessianOfFlow = hessian;
   partialDerivatives = NULL;
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet);
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::operator()(T& originalSet, CnCoeffType& result)
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = NULL;
   hessianOfFlow = NULL;
   partialDerivatives = &result;
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return result()=crossSection<T>(originalSet);
}

// -------------------- protected functions ---------------------------------

// this function is just for compatibility with C^1 and C^2 routines
// see the version for C^1-set types below
template <typename DS, typename FunT>
inline void PoincareMap<DS, FunT>::computeJacEnclosure(C0Set& S, VectorType& V)
{
   
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT>
void PoincareMap<DS, FunT>::computeJacEnclosure(C1Set& S, VectorType& V)
{
   if(derivativeOfFlow!=NULL)
   {
      MatrixType enc = dynamicalSystem.jacEnclosure(V,*(S.getNorm()));
      MatrixType oneStepBound = enc*MatrixType(S);
      for(int i=0;i<m_dim;i++)
         for(int j=0;j<m_dim;j++)
            (*derivativeOfFlow)[i][j] = intervalHull((*derivativeOfFlow)[i][j],oneStepBound[i][j]);
   }
   
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT>
void PoincareMap<DS, FunT>::computeJacEnclosure(C2Set& S, VectorType& enc)
{
   if(derivativeOfFlow!=NULL)
   {
      int i,j,c;
      MatrixType jacEnc = dynamicalSystem.jacEnclosure(enc,*(S.getNorm()));
      MatrixType currentDerivative = MatrixType(S);

// computation of the hessian of Poincare map

      if(hessianOfFlow!=NULL)
      {
         MatrixType* oneStepHessian = new (m_dim,m_dim) MatrixType[m_dim];
         typename DS::C2CoeffType EH(m_dim); // EH will be an enclosure for the hessian part
         dynamicalSystem.c2Enclosure(enc,jacEnc,*(S.getNorm()),EH);
         int ss,r;

         for(i=0;i<m_dim;++i)
            oneStepHessian[i].clear();

         for(i=0;i<m_dim;++i)
            for(j=0;j<m_dim;++j)
               for(c=0;c<m_dim;++c)
                  for(ss=0;ss<m_dim;++ss)
                     for(r=0;r<m_dim;++r)
                        //oneStepHessian[i][j][c] += EH(i,ss,r) * (*derivativeOfFlow)[ss][j] * (*derivativeOfFlow)[r][c];
                        oneStepHessian[i][j][c] += EH(i,ss,r) * currentDerivative[ss][j] * currentDerivative[r][c];

         for(i=0;i<m_dim;++i)
            for(j=0;j<m_dim;++j)
               for(c=0;c<m_dim;++c)
                  for(ss=0;ss<m_dim;++ss)
                     oneStepHessian[i][j][c] += jacEnc[i][ss] * S(ss,j,c)/*hessianOfFlow[ss][j][c]*/;

         for(i=0;i<m_dim;i++)
            for(j=0;j<m_dim;j++)
               for(c=0;c<m_dim;c++)
                  hessianOfFlow[i][j][c] = intervalHull(hessianOfFlow[i][j][c],oneStepHessian[i][j][c]);

         delete []oneStepHessian;
      } // end if(hessianOfFlow!=NULL)

// computation of enclosure for derivative of the Poincare map
      MatrixType oneStepBound = jacEnc*currentDerivative;
      for(i=0;i<m_dim;i++)
         for(j=0;j<m_dim;j++)
            (*derivativeOfFlow)[i][j] = intervalHull((*derivativeOfFlow)[i][j],oneStepBound[i][j]);
   } // end of if(derivativeOfFlow!=NULL)
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT>
void PoincareMap<DS, FunT>::computeJacEnclosure(CnSet& S, VectorType& )
{
   if(partialDerivatives!=NULL)
   {
      CnCoeffType enc(m_dim,partialDerivatives->rank()),
                  oneStepBound(m_dim,partialDerivatives->rank());

      dynamicalSystem.cnEnclosure(VectorType(S),enc,*(S.getNorm()));
      //dynamicalSystem.computeComposition(S.currentSet(),enc,oneStepBound);
      substitutionPowerSeries(enc,S.currentSet(),oneStepBound,false);

      for(int p=0;p<m_dim;++p)
      {
         typename CnCoeffType::iterator b=partialDerivatives->begin(p,1), e=partialDerivatives->end(p,partialDerivatives->rank());
         typename CnCoeffType::iterator i=oneStepBound.begin(p,1);
         while(b!=e)
         {
            *b = intervalHull(*b,*i);
            ++b;
            ++i;
         }
      }
   }
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT>
void PoincareMap<DS, FunT>::saveMatrices(CnSet& S)
{
   if(partialDerivatives!=NULL) *partialDerivatives = S.currentSet();
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT>
void PoincareMap<DS, FunT>::saveMatrices(C2Set& S)
{
   if(derivativeOfFlow!=NULL)
      *derivativeOfFlow = MatrixType(S);

   if(hessianOfFlow!=NULL)
   {
      int i,j,c;
      for(i=0;i<m_dim;++i)
         for(j=0;j<m_dim;++j)
            for(c=0;c<m_dim;++c)
               hessianOfFlow[i][j][c] = S(i,j,c);
   }
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT>
void PoincareMap<DS, FunT>::saveMatrices(C1Set& S)
{
   if(derivativeOfFlow!=NULL) 
     *derivativeOfFlow = MatrixType(S);
}

/*__________________________________________________________________________*/

template <typename DS, typename FunT> template<typename T>
void PoincareMap<DS, FunT>::reachSection(T& theSet)
// The common procedure to reach section for all types of the sets.
// As a result we obtain the set closer to section than sectionFactor * sizeOfSet
{
  ScalarType step = dynamicalSystem.getStep();// time step used to reach the section
  ScalarType tmp;
  VectorType beforeSection;
  VectorType bound;
  

  VectorType startSet = VectorType(theSet);
  
  // We want to leave section and reach point where the sign of the section function,
  // according to crossingDirection, indicate that next section crossing will be in a good direction.
  while(true)
  {
    ScalarType sign = section(startSet);
     
    if(sign.contains(0.0) )
    {
     // we verify if we can leave the section, i.e. if the inner product
     // of the vectorFiled and the normal vector to the section is non-zero 
      VectorType startEnclosure = dynamicalSystem.enclosure(startSet);
      ScalarType innerProduct = section.gradient(startEnclosure)*dynamicalSystem.getField()(startEnclosure);
      if(innerProduct.contains(0.0))
      {
         throw PoincareException<T>(
            "Poincare section error: cannot leave the section",theSet,theSet,
            startEnclosure, dynamicalSystem.getField()(startEnclosure),
            section(startEnclosure), section.gradient(startEnclosure), innerProduct
         );
      }
    }
    else 
      if((crossingDirection == None) || 
         ((crossingDirection == MinusPlus) && (sign < 0.0)) ||
         ((crossingDirection == PlusMinus) && (sign > 0.0))) 
        break;

      theSet.move(dynamicalSystem); // we make one step to leave the section and try again
      startSet = VectorType(theSet);
   }

// used for checking if we cross the section (if sign is changed)
   ScalarType check, sign;

// first return to section with constant time step
   beforeSection = startSet; // in fact it is equal to VectorType(theSet);
   check = section(beforeSection);
   sign = check.mid(); // we save the sign of the section function, check do not contain zero
   T temp(theSet);
   T *original = &theSet, *swapSet, *tempSet = &temp;

// now we try to move the set closer to section than one current time step
   while(check*sign>0)
   {
      bound = dynamicalSystem.enclosure(beforeSection);
      original->move(dynamicalSystem,*tempSet); // bound is enclosure for whole time step
      tmp = section.gradient(bound)*dynamicalSystem.getField()(bound);
      if (tmp.contains(0.0))
      {  // we check if we cross the section and than return during one step
         check = section(bound);
         if(subset(ScalarType(0.0),check))
         {
            throw PoincareException<T>(
               "Poincare section error:",theSet,temp,
               bound, dynamicalSystem.getField()(bound),
               check, section.gradient(bound), tmp
            );
         }
         beforeSection = VectorType(*tempSet);
      }
      else
      {
         beforeSection = VectorType(*tempSet);
         check = section(beforeSection);
      }
      swapSet = original;
      original = tempSet;
      tempSet = swapSet;
  } // end while

   if(&theSet != tempSet) theSet = *tempSet;

// We want to come to section closer than  sectionFactor * size of set in section gradient direction

   beforeSection = VectorType(theSet);
   VectorType Grad = section.gradient(beforeSection);

   typename ScalarType::BoundType maxDistance = diam(abs(beforeSection*Grad)).rightBound()*sectionFactor;
   int maxIterations = 30; // usually 2-3 iterations are necessary to come close to section

   original = &theSet;
   tempSet = &temp;

   while(maxIterations)
   {
      VectorType Mid = midVector(beforeSection);
      VectorType fieldDirection = dynamicalSystem.getField()(Mid);
      ScalarType approxDistance = (abs(section(beforeSection))).leftBound();

      if(approxDistance.leftBound() < maxDistance)      // we are close enough
         break;

      ScalarType fieldSectionDirection = abs(fieldDirection * Grad);
      ScalarType approxTime = ((approxDistance-(maxDistance/15))/ fieldSectionDirection).mid();

      if (step<0)
         approxTime = -approxTime;

      dynamicalSystem.setStep(approxTime.leftBound());
      bound = dynamicalSystem.enclosure(beforeSection);
      tmp = section.gradient(bound)*dynamicalSystem.getField()(bound);
      original->move(dynamicalSystem,*tempSet);

      while(true)
      {
         if (tmp.contains(0.0)) // we check if we cross the section and than return during one step
         {
            check = section(bound);
            if(subset(ScalarType(0.),check))
            {
               throw PoincareException<T>(
                  "Poincare section gradient error:", theSet, temp,
                  bound, dynamicalSystem.getField()(bound),
                  check, section.gradient(bound), tmp
               );
            }
         }
         else
            check = section(VectorType(*tempSet));
         if(!(check*sign>0))
         {
            dynamicalSystem.setStep(dynamicalSystem.getStep() * timeStepCorrectionFactor);
            bound = dynamicalSystem.enclosure(beforeSection);
            tmp = section.gradient(bound)*dynamicalSystem.getField()(bound);
            original->move(dynamicalSystem,*tempSet);
         }        
         else 
           break;
      } // end while

      swapSet = original;
      original = tempSet;
      tempSet = swapSet;

      beforeSection = VectorType(*original);
      Grad = section.gradient(beforeSection);
      maxIterations--;
   }

   dynamicalSystem.setStep(step);
   if(&theSet != original) 
     theSet = *original;
}

/*__________________________________________________________________________*/

// the following function is used to cross the section for all types of sets

template <typename DS, typename FunT> 
template<typename T>
typename PoincareMap<DS, FunT>::VectorType PoincareMap<DS, FunT>::crossSection(T& theSet)
{
// theSet is very close to the section
// we compute an approximate time needed to cross the section
// if the approximate values of time step is to small to change the sign of the section function
// the 'smallTime' time is used to move the set
   ScalarType step = dynamicalSystem.getStep();
  
  //int maxIterations = 30; // usually 2-3 iterations are necessary to come close to section

  // We want stop after section closer than  maxDistance =  sectionFactor * size of set in section gradient direction 
  VectorType Set = VectorType(theSet);
  VectorType result = Set;
  VectorType Grad = section.gradient(Set);
  typename ScalarType::BoundType maxDistance = diam(Set*Grad).rightBound()*sectionFactor;
  typename ScalarType::BoundType approxDistance = - abs(section(Set)).rightBound(); 
  
  ScalarType check = section(Set);
  if(check.contains(0.0)){
    throw PoincareException<T>("Poincare exception in section crossing : initial set is already on the section ", theSet, check );
  }
  ScalarType sign = mid(check);
  
  
  // direction of section crossing =  PlusMinus or MinusPlus
  // we check sign of the section instead of checking value of variable crossingDirection because it can be set to None 
  typename BasicPoincareMap<DS, FunT>::CrossingDirection direction = ( check > 0)? PlusMinus: MinusPlus;
  
  T temp(theSet); // we save the set before section, used in throwing exception only
  saveMatrices(theSet); // we save the \frac{d}{dx}\phi

  T* original = &theSet;
  T* tempSet = &temp;
  
  while(!(check*sign < 0)){
    
    VectorType fieldDirection = dynamicalSystem.getField()(midVector(Set));
    ScalarType fieldSectionDirection = abs(fieldDirection * Grad);
    //ScalarType approxTime = capd::min(((maxDistance/2 - approxDistance ) / fieldSectionDirection).mid(), abs(step));
    ScalarType approxTime = ((maxDistance/2 - approxDistance ) / fieldSectionDirection).mid();
//    std::cout << "\n---\n maxDistance = " << maxDistance << "\n approxDistance = " << approxDistance<< "\n approxTime = " << approxTime;    
    approxTime = capd::min(approxTime, abs(step));
    
    if (step<0)
      approxTime = -approxTime;
    
    while(true){
      dynamicalSystem.setStep(approxTime.left());
      original->move(dynamicalSystem,*tempSet);
      
      Set = VectorType(*tempSet);
      if(direction == PlusMinus)
        approxDistance = -(section(Set).rightBound());
      else
        approxDistance = section(Set).leftBound();
      
      if(approxDistance  > maxDistance) 
        approxTime *= timeStepCorrectionFactor;
      else
        break;
    }
      
    VectorType bound = dynamicalSystem.enclosure(VectorType(*original));
    
    computeJacEnclosure(theSet,bound); 
    VectorType Fresult = dynamicalSystem.getField()(bound);
    if(subset(ScalarType(0.),section.gradient(bound)*Fresult)){ // if the vector field is tangent to the section
      VectorType secGradient = section.gradient(bound);
      throw PoincareException<T>(
          "Poincare section gradient error:", theSet, temp,
          bound, Fresult, check, secGradient, secGradient * Fresult
      );
    }
    for(int i=0;i<m_dim;i++){
      if(!(Fresult[i].contains(0.)))
         result[i] = intervalHull(result[i],Set[i]);
      else
         result[i] = intervalHull(result[i],bound[i]); // bound is an enclosure
    }
    T* swapSet = original;
    original = tempSet;
    tempSet = swapSet;
    
    check = section(Set); // we compute the sign of the section
  } // end while

  dynamicalSystem.setStep(step);
  
  if(&theSet != original) 
    theSet = *original;

  return result;
}

}} // namespace capd::poincare

#endif // _CAPD_POINCARE_POINCARE_MAP_HPP_

/// @}
