/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Parser.cpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#include <string>
#include <stdexcept>
#include <list>
#include "map/Parser.h"

using namespace capd::map;

/* _________________________ parseBrackets ____________________________*/

/**
   return a position of '(' which is paired with the last ')'
   or 0, when there are not any parentheses
*/

size_t Parser::parseBrackets(const std::string &e, size_t position)
{
	size_t len = e.find_last_of(')', position);
	if (len==std::string::npos) return 0; //  no parenthese found

	size_t cnt=1;  // contains a difference of the numberes of ')' and '('

	do{
		do{
			if(len==0)
			{
				std::string message = "Parse error in the formula ";
				message += e;
				throw std::runtime_error(message);
			} 
			len--;
		}
		while((e.at(len)!='(')&&(e.at(len)!=')'));

		if(e.at(len)==')') cnt++;
		if(e.at(len)=='(')	cnt--;

	}while(cnt);
	return len;
}

/* _________________________ searchForOperator _____________________ */

size_t Parser::searchForOperator(const std::string& e, unsigned char op, size_t position)
/*
   checks if operator <op> appears in text <e> outside any parentheses
   and returns its position
   if not 0 is returned

   the search starts at the end
*/
{
	size_t isOperator = e.find_last_of(op,position);
	if (isOperator == std::string::npos) return 0; // no operator found

	size_t lastBracket = e.find_last_of(')',position);
	if(lastBracket == std::string::npos || isOperator>lastBracket) return isOperator; // no brackets found

 	size_t pB = parseBrackets(e,position);
	if(pB) return searchForOperator(e,op,pB-1);
	return 0;
}


/* ________________________ searchForFunction _______________________ */

size_t Parser::searchForFunction(const std::string &fun, const std::string &e)
/**
   returns position of arguments of function <fun> , or
   0 if <fun> does not appaer or is an argument for some other function
*/
{
	size_t pos = e.find(fun);
	if(pos) return 0;
	return parseBrackets(e);
}


/* __________________________ removeBrackets ______________________ */

std::string & Parser::removeBrackets(std::string &eq)
/**
	This function removes exterior brackets
*/
{
	while ((!parseBrackets(eq))&&(eq.at(eq.size()-1)==')'))
 	{
		eq.erase(eq.size()-1);
		eq.erase(0,1);
 	}
	return eq;
}

/*---------------------- splitVariables -----------------------------*/

/*
 This function searches for text 'what' in 'where', then cuts simple words
 separated by: ':' ',' ';' ' ' begining from word next to 'what'.
 The last word is indicated by ';' after that word.
 These words are written to the table tab[].
*/

std::string *Parser::splitVariables(const std::string &what, const std::string &where, int &ile)
{
	std::list<std::string> temp;

	size_t start = where.find(what);
	if(start==std::string::npos)
	{
		std::string message = "Connot find '";
		message += what;
		message += "' in ";
		message += where;
		throw std::runtime_error(message);
	}
	start = where.find_first_of(':',start);
	size_t last = where.find_first_of(';',start);
	
	ile = 0;
	while(start<last)
	{
		size_t p = where.find_last_of(", :;",last-1);
		if(p==std::string::npos) break;
		
		temp.push_front(where.substr(p+1,last-p-1));
		++ile;
		last=p;
	}
	std::string *tab = new std::string[ile];
	std::copy(temp.begin(),temp.end(),tab);
	return tab;
}

/// @}
